//6
////////////////////////////////////////////////////////
// twaka
// written by Andrew Johnson - Summer 2007
// making use of some of the evl cave code
// evl cave code is copyright board of regents, U of I
// aej@evl.uic.edu
////////////////////////////////////////////////////////
// twaka was written as a substitute for real tracker data
// while working on VR code that makes use of trackd
////////////////////////////////////////////////////////
// by default the tracker sends data in feet and degrees
// 1 foot = 0.305 meters

// need to try this under windows
// init controller data and init tracker data

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifdef _WIN32
typedef unsigned int uint32_t;
typedef int int32_t;
#else
#include <sys/shm.h>
#endif

#include <signal.h>
#include <limits.h>

#include "trackd.h"

#ifdef _WIN32
#include <windows.h>
#endif

#ifdef linux
#include <GL/glut.h>
#else
#include "glut.h"
#endif

#define PERMS 0666

#define CONTROLLER_KEY 4127
#define TRACKER_KEY 4126

#define WIN_WIDTH 500
#define WIN_HEIGHT 250

#define FRAME_RATE 30


#ifndef _WIN32
static int controller_shmid = -1;
static int tracker_shmid = -1; 
#else
static volatile HANDLE tracker_shmid = NULL;
static volatile HANDLE controller_shmid = NULL;
#endif


static struct TRACKD_CONTROLLER	* controller = NULL;
struct TRACKD_TRACKING	* tracker = NULL;

int win1;
	
////////////////////////////////////////////////////////////

#ifndef DVORAK

#define k_JOY_NX 'a'
#define k_JOY_PX 'd'
#define k_JOY_NY 's'
#define k_JOY_PY 'w'
#define K_JOY_NX 'A'
#define K_JOY_PX 'D'
#define K_JOY_NY 'S'
#define K_JOY_PY 'W'

#define k_WAND_XP 'l'
#define k_WAND_XN 'j'
#define k_WAND_YP 'i'
#define k_WAND_YN 'k'
#define k_WAND_ZP 'm'
#define k_WAND_ZN 'u'

#define k_WAND_ELEVP 'I'
#define k_WAND_ELEVN 'K'
#define k_WAND_AZIMP 'J'
#define k_WAND_AZIMN 'L'

#define k_HEAD_XP 'h'
#define k_HEAD_XN 'f'
#define k_HEAD_YP 't'
#define k_HEAD_YN 'g'
#define k_HEAD_ZP 'v'
#define k_HEAD_ZN 'r'

#define k_QUIT 'q'
#define K_QUIT 'Q'

#else

#define k_JOY_NX 'a'
#define k_JOY_PX 'e'
#define k_JOY_NY 'o'
#define k_JOY_PY ','
#define K_JOY_NX 'A'
#define K_JOY_PX 'E'
#define K_JOY_NY 'O'
#define K_JOY_PY '<'

#define k_WAND_XP 'n'
#define k_WAND_XN 'h'
#define k_WAND_YP 'c'
#define k_WAND_YN 't'
#define k_WAND_ZP 'm'
#define k_WAND_ZN 'g'

#define k_WAND_ELEVP 'C'
#define k_WAND_ELEVN 'T'
#define k_WAND_AZIMP 'H'
#define k_WAND_AZIMN 'N'

#define k_HEAD_XP 'd'
#define k_HEAD_XN 'u'
#define k_HEAD_YP 'y'
#define k_HEAD_YN 'i'
#define k_HEAD_ZP 'k'
#define k_HEAD_ZN 'p'

#define k_QUIT '\''
#define K_QUIT '"'

#endif

////////////////////////////////////////////////////////////

void display(void);
	
////////////////////////////////////////////////////////////
// void trackd_cleanup()
//	When the daemon is interrupted by SIGINT, this routine closes the
//   connections, disposes of the memory, and exits.
////////////////////////////////////////////////////////////

void trackd_cleanup(void)
{
#ifndef _WIN32
struct shmid_ds *buf; // added by andy
#endif

 fprintf(stderr,"Quitting...\n");
 
#ifndef _WIN32
 if (controller)
	if (shmdt(controller) < 0)
		perror("Detaching 'controller' shared memory");

 if (tracker)
	if (shmdt(tracker) < 0)
		perror("Detaching 'tracker' shared memory");

		
 if (controller_shmid > -1)
	if (shmctl(controller_shmid,IPC_RMID, buf) < 0)
		perror("Removing 'controller_shmid'");

 if (tracker_shmid > -1)
	if (shmctl(tracker_shmid,IPC_RMID, buf) < 0)
		perror("Removing 'tracker_shmid'");
#else
 if (controller)
	UnmapViewOfFile(controller);
    perror("Detaching 'controller' shared memory");

 if (tracker)
	UnmapViewOfFile(tracker);
    perror("Detaching 'tracker' shared memory");

#endif
		
 exit(-1);
}


////////////////////////////////////////////////////////////
// void get_controller_mem()
//	Allocates shared memory for the buttons/joystick
//  The global 'controller' points to the chunk
////////////////////////////////////////////////////////////

void get_controller_mem(int ControllerdKey)
    {
    char shmkey[256];
    
    fprintf(stderr, "Controller shared memory requested: %d ", sizeof(struct TRACKD_CONTROLLER));

#ifndef _WIN32
    controller_shmid = shmget(ControllerdKey, 
			sizeof(struct TRACKD_CONTROLLER), PERMS | IPC_CREAT);
			
    if (controller_shmid < 0)
        {
        fprintf(stderr,"can't get shared memory\n");
        exit(-1);
        }
    
    controller = (struct TRACKD_CONTROLLER *) shmat(controller_shmid,(char *) 0, 0);
            
    if (controller == (struct TRACKD_CONTROLLER *) -1)
        {
        fprintf(stderr,"can't attach shared memory\n");
        exit(-1);
        }
#else

    sprintf(shmkey, "%d", ControllerdKey);
    controller_shmid = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE,0,
		sizeof(struct TRACKD_CONTROLLER), shmkey);

    if (controller_shmid == NULL)
        {
        fprintf(stderr,"can't get shared memory\n");
        exit(-1);
        }    
    
    controller = (struct TRACKD_CONTROLLER *) MapViewOfFile(controller_shmid, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(struct TRACKD_CONTROLLER));
    if (controller == NULL)
        {
        fprintf(stderr,"can't attach shared memory\n");
        exit(-1);
        }
#endif        
        
    fprintf(stderr, " <done>\n");
    }

////////////////////////////////////////////////////////////
// void get_tracker_mem()
////////////////////////////////////////////////////////////

void get_tracker_mem(int TrackerdKey)
    {
    char shmkey[256];
    
    fprintf(stderr, "Tracker shared memory requested: %d ", sizeof(struct TRACKD_TRACKING));

#ifndef _WIN32    
    tracker_shmid = shmget(TrackerdKey, 
                sizeof(struct TRACKD_TRACKING), PERMS | IPC_CREAT);
                
    if (tracker_shmid < 0)
        {
        fprintf(stderr,"can't get shared memory for tracker\n");
        exit(-1);
        }
        
    tracker = (struct TRACKD_TRACKING *) shmat(tracker_shmid,(char *) 0, 0);
                
    if (tracker == (struct TRACKD_TRACKING *) -1)
        {
        fprintf(stderr,"can't attach shared memory for tracker\n");
        exit(-1);
        }
#else

    sprintf(shmkey, "%d", TrackerdKey);
    tracker_shmid = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE,0,
		sizeof(struct TRACKD_TRACKING), shmkey);

    if (tracker_shmid == NULL)
        {
        fprintf(stderr,"can't get shared memory\n");
        exit(-1);
        }    
    
    tracker = (struct TRACKD_TRACKING *) MapViewOfFile(tracker_shmid, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(struct TRACKD_TRACKING));
    if (tracker == NULL)
        {
        fprintf(stderr,"can't attach shared memory\n");
        exit(-1);
        }
#endif        

    fprintf(stderr, " <done>\n");
    }

////////////////////////////////////////////////////////////
// void init_controller_data()
//	Fills in the controller header and data with initial values.
////////////////////////////////////////////////////////////

void init_controller_data(void)
    {
    int loop;
    
    controller->header.version         = CAVELIB_2_6;
    controller->header.buttonOffset    = ((char *)&controller->controller.button[0]) -
                                         ((char *)controller);
    controller->header.valuatorOffset  = ((char *)&controller->controller.valuator[0]) -
                                         ((char *)controller);
    controller->header.numButtons      = 3;
    controller->header.numValuators    = 2;
    controller->header.timestamp[0]    = controller->header.timestamp[1] = 0;
    controller->header.command         = 0;
    
    for (loop=0; loop < CAVE_MAX_BUTTONS; ++loop) 
        controller->controller.button[loop] = 0;
        
    for (loop=0; loop < CAVE_MAX_VALUATORS; ++loop) 
        controller->controller.valuator[loop] = 0.0;
        
    controller->controller.num_buttons = 3;
    controller->controller.num_valuators = 2;
    }

////////////////////////////////////////////////////////////

void init_tracker_data(void)
    {
    int loop;
    
    tracker->header.version         = CAVELIB_2_6;
    tracker->header.numSensors      = TRACKD_MAX_SENSORS;
    tracker->header.sensorOffset    = ((char *)&tracker->sensor[0]) -
                                       ((char *)tracker);
    tracker->header.sensorSize      = sizeof(CAVE_SENSOR_ST);
    tracker->header.timestamp[0]    = tracker->header.timestamp[1] = 0;
    tracker->header.command         = 0;
    
    for (loop=0; loop < TRACKD_MAX_SENSORS; ++loop) 
        {
        tracker->sensor[loop].x = 0;
        tracker->sensor[loop].y = 0;
        tracker->sensor[loop].z = 0;
    
        tracker->sensor[loop].azim = 0;
        tracker->sensor[loop].elev = 0;
        tracker->sensor[loop].roll = 0;
    
        tracker->sensor[loop].calibrated = 0; // not sure what this should be
        tracker->sensor[loop].frame = CAVE_TRACKER_FRAME; /* CAVE_TRACKER_FRAME or CAVE_NAV_FRAME */
        }
    
    // place the head 5 feet off the floor by default
    tracker->sensor[0].x = 0;
    tracker->sensor[0].y = 5;
    tracker->sensor[0].z = 0;
    
    
    // place the wand relative to the head
    tracker->sensor[1].x = tracker->sensor[0].x + 1;
    tracker->sensor[1].y = tracker->sensor[0].y - 1;
    tracker->sensor[1].z = tracker->sensor[0].z - 3;
    }

////////////////////////////////////////////////////////////

void output(int x, int y, char *string)
    {
    char *c;

    glColor3f(0.0, 0.0, 0.0);
    
    glRasterPos2f(x, y);
    for (c=string; *c != '\0'; c++)
        glutBitmapCharacter(GLUT_BITMAP_9_BY_15 , *c);

    glEnable(GL_DEPTH_TEST);
    }


////////////////////////////////////////////////////////////
// recompute the field for the given locations
    
void update(int value)
    {
    glutTimerFunc(FRAME_RATE, update, 1);
    } 
    
////////////////////////////////////////////////////////////

void idle(void)
    {
#if defined(WIN32) || defined(_WIN32)
    Sleep(10);
#else
    //usleep(10000);
    usleep(100000);
#endif

    display();
    }
    
////////////////////////////////////////////////////////////

void mouse ( int button, int state, int x, int y )
    {            
    }

////////////////////////////////////////////////////////////

void motion ( int x, int y )
    {
    }

////////////////////////////////////////////////////////////

#define TRACKER_SPEED 0.05

void keyboard(unsigned char c, int x, int y)
    {
    switch(c) {
    
        // joystick
        case k_JOY_NX:
        case K_JOY_NX: controller->controller.valuator[0] = -0.5; break;
        
        case k_JOY_PX:
        case K_JOY_PX: controller->controller.valuator[0] = 0.5; break;
        
        case k_JOY_NY:
        case K_JOY_NY: controller->controller.valuator[1] = -0.5; break;
    
        case k_JOY_PY:
        case K_JOY_PY: controller->controller.valuator[1] = 0.5; break;
        
    
        // moving the user (ie simulated walking)
        case k_HEAD_YP: tracker->sensor[0].y += TRACKER_SPEED; break;
        case k_HEAD_YN: tracker->sensor[0].y -= TRACKER_SPEED; break;
        case k_HEAD_XN: tracker->sensor[0].x -= TRACKER_SPEED; break;
        case k_HEAD_XP: tracker->sensor[0].x += TRACKER_SPEED; break;
        case k_HEAD_ZN: tracker->sensor[0].z -= TRACKER_SPEED; break;
        case k_HEAD_ZP: tracker->sensor[0].z += TRACKER_SPEED; break;
    
        // turning the user's head
        // not visible on a single screen projection system
    
    
        // moving the wand
        case k_WAND_YP: tracker->sensor[1].y += TRACKER_SPEED; break;
        case k_WAND_YN: tracker->sensor[1].y -= TRACKER_SPEED; break;
        case k_WAND_XN: tracker->sensor[1].x -= TRACKER_SPEED; break;
        case k_WAND_XP: tracker->sensor[1].x += TRACKER_SPEED; break;
        case k_WAND_ZN: tracker->sensor[1].z -= TRACKER_SPEED; break;
        case k_WAND_ZP: tracker->sensor[1].z += TRACKER_SPEED; break;
    
         //turning the wand
        case k_WAND_ELEVP:
            tracker->sensor[1].elev += 1;
            if (tracker->sensor[1].elev > 89)
                tracker->sensor[1].elev = 89;
            break;
            
        case k_WAND_ELEVN:
            tracker->sensor[1].elev -= 1;
            if (tracker->sensor[1].elev < -89)
                tracker->sensor[1].elev = -89;
            break;
            
        case k_WAND_AZIMP:
            tracker->sensor[1].azim += 1;
            if (tracker->sensor[1].azim > 179)
                tracker->sensor[1].azim = 179;
            break;
            
        case k_WAND_AZIMN:
            tracker->sensor[1].azim -= 1;
            if (tracker->sensor[1].azim < -179)
                tracker->sensor[1].azim = -179;
            break;
            
        // buttons on the wand
        case '1': controller->controller.button[0] = 1; break;
        case '2': controller->controller.button[1] = 1; break;
        case '3': controller->controller.button[2] = 1; break;
    
        // quitting twaka
        case K_QUIT:
        case k_QUIT: trackd_cleanup(); break;
        }
    }

////////////////////////////////////////////////////////////

void keyboardUp(unsigned char c, int x, int y)
    {
    switch(c) {
        case k_JOY_NX:
        case K_JOY_NX:
        case k_JOY_PX:
        case K_JOY_PX: controller->controller.valuator[0] = 0; break;
        
        case k_JOY_NY: 
        case K_JOY_NY: 
        case k_JOY_PY:
        case K_JOY_PY: controller->controller.valuator[1] = 0; break;

        case '1': controller->controller.button[0] = 0; break;
        case '2': controller->controller.button[1] = 0; break;
        case '3': controller->controller.button[2] = 0; break;
        }
    }

////////////////////////////////////////////////////////////

void display(void)
    {
    char line[255];
    
    glClearColor(1.0, 1.0, 1.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, WIN_WIDTH, 0, WIN_HEIGHT, -10, 10);
	glScalef(1, -1, 1);
	glTranslatef(0, -WIN_HEIGHT, 0);
	glMatrixMode(GL_MODELVIEW);


    output(10, 20,   "1  2  3  - wand buttons");
    
    output(10, 60,   "   w");
    output(10, 75,   "a  s  d  - wand joystick");
    
    output(10, 115,   "u  i");
    output(10, 130,   "j  k  l  - wand movement");
    output(10, 145,  "m  ");
    
    
    output(275, 20,  "  q/Q    - quit");
        
    output(275, 60,  "r  t");
    output(275, 75,  "f  g  h  - head movement");
    output(275, 90,  "v  ");

    output(275, 115,  "   I");
    output(275, 130,  "J  K  L  - wand turning");

    output(10, 165, "-----------------------------------------------------");

    // here we should print out the current tracker values
    sprintf(line, "head:  %5.1f %5.1f %5.1f   - %4.0f %4.0f %4.0f", 
        tracker->sensor[0].x, tracker->sensor[0].y, tracker->sensor[0].z,
        tracker->sensor[0].azim, tracker->sensor[0].elev, tracker->sensor[0].roll);
    output(10,180, line);

    sprintf(line, "wand:  %5.1f %5.1f %5.1f   - %4.0f %4.0f %4.0f", 
        tracker->sensor[1].x, tracker->sensor[1].y, tracker->sensor[1].z,
        tracker->sensor[1].azim, tracker->sensor[1].elev, tracker->sensor[1].roll);
    output(10,200, line);

    sprintf(line, "buttons:  %d %d %d            -    joystick %4.1f %4.1f",
        controller->controller.button[0], controller->controller.button[1], controller->controller.button[2],
        controller->controller.valuator[0], controller->controller.valuator[1]);
    output(10,230, line);

    
    glutSwapBuffers();
    }

////////////////////////////////////////////////////////////
 
void dealWithParameters(int argc, char **argv)
    {
    }

////////////////////////////////////////////////////////////

void reshape(int width, int height)
    {
	// Set the viewport to be the entire window
    glViewport(0, 0, width, height); 
    }
    
////////////////////////////////////////////////////////////

main(int argc, char **argv)
    {
    glutInit(&argc, argv);
    
    dealWithParameters(argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
        
#if defined(WIN32) || defined(_WIN32)
    glutInitWindowPosition(5, 30);
#else
    glutInitWindowPosition(0, 0);
#endif
    
    glutInitWindowSize(WIN_WIDTH, WIN_HEIGHT);

    win1 = glutCreateWindow("Electro tracker simulator");

    if (glutGet(GLUT_WINDOW_COLORMAP_SIZE) != 0)
	   {
	   printf("FAIL: bad RGBA colormap size\n");
	   exit(1);
	   }

    glutKeyboardFunc(keyboard);
    glutKeyboardUpFunc(keyboardUp);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);

    glutReshapeFunc(reshape);

    glutDisplayFunc(display);
    glutIdleFunc(idle);
    glutTimerFunc(FRAME_RATE, update, 1);


    get_controller_mem(CONTROLLER_KEY);
    init_controller_data();
    get_tracker_mem(TRACKER_KEY);
    init_tracker_data(); 

    glutMainLoop();
    return 0;  
    }

